import unittest
import json

import six

import pynetbox
from .util import Response

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


api = pynetbox.api(
    "http://localhost:8000",
    token="abc123",
)

nb = api.ipam

HEADERS = {
    'accept': 'application/json;',
    'authorization': 'Token abc123',
}

POST_HEADERS = {
    'Content-Type': 'application/json;',
    'authorization': 'Token abc123',
}


class Generic(object):
    class Tests(unittest.TestCase):
        name = ''
        name_singular = None
        ret = pynetbox.core.response.Record
        app = 'ipam'

        def test_get_all(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name
                ))
            ) as mock:
                ret = getattr(nb, self.name).all()
                self.assertTrue(ret)
                self.assertTrue(isinstance(ret, list))
                self.assertTrue(isinstance(ret[0], self.ret))
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )

        def test_filter(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name
                ))
            ) as mock:
                ret = getattr(nb, self.name).filter(name='test')
                self.assertTrue(ret)
                self.assertTrue(isinstance(ret, list))
                self.assertTrue(isinstance(ret[0], self.ret))
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={"name": "test"},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )

        def test_get(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name_singular or self.name[:-1]
                ))
            ) as mock:
                ret = getattr(nb, self.name).get(1)
                self.assertTrue(ret)
                self.assertTrue(isinstance(ret, self.ret))
                self.assertTrue(isinstance(dict(ret), dict))
                self.assertTrue(isinstance(str(ret), str))
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/1/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )


class PrefixTestCase(Generic.Tests):
    name = 'prefixes'
    name_singular = 'prefix'

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='ipam/prefix.json')
    )
    def test_modify(self, *_):
        ret = nb.prefixes.get(1)
        ret.prefix = '10.1.2.0/24'
        ret_serialized = ret.serialize()
        self.assertTrue(ret_serialized)
        self.assertFalse(ret._compare())
        self.assertEqual(ret_serialized['prefix'], '10.1.2.0/24')

    @patch(
        'pynetbox.core.query.requests.sessions.Session.put',
        return_value=Response(fixture='ipam/prefix.json')
    )
    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='ipam/prefix.json')
    )
    def test_idempotence(self, *_):
        ret = nb.prefixes.get(1)
        test = ret.save()
        self.assertFalse(test)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        side_effect=[
            Response(fixture='ipam/prefix.json'),
            Response(fixture='ipam/available-ips.json'),
        ]
    )
    def test_get_available_ips(self, mock):
        pfx = nb.prefixes.get(1)
        ret = pfx.available_ips.list()
        mock.assert_called_with(
            'http://localhost:8000/api/ipam/prefixes/1/available-ips/',
            params={},
            json=None,
            headers=HEADERS,
            verify=True
        )
        self.assertTrue(ret)
        self.assertEqual(len(ret), 3)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.post',
        return_value=Response(fixture='ipam/available-ips-post.json')
    )
    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='ipam/prefix.json'),
    )
    def test_create_available_ips(self, _, post):
        expected_result = {
            'status': 1,
            'description': '',
            'nat_inside': None,
            'role': None,
            'vrf': None,
            'address':
            '10.1.1.1/32',
            'interface': None,
            'id': 1,
            'tenant': None
        }
        create_parms = dict(
            status=2,
        )
        pfx = nb.prefixes.get(1)
        ret = pfx.available_ips.create(create_parms)
        post.assert_called_with(
            'http://localhost:8000/api/ipam/prefixes/1/available-ips/',
            params={},
            headers=POST_HEADERS,
            json=create_parms,
            verify=True
        )
        self.assertTrue(ret)
        self.assertEqual(ret, expected_result)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        side_effect=[
            Response(fixture='ipam/prefix.json'),
            Response(fixture='ipam/available-prefixes.json'),
        ]
    )
    def test_get_available_prefixes(self, mock):
        pfx = nb.prefixes.get(1)
        ret = pfx.available_prefixes.list()
        mock.assert_called_with(
            'http://localhost:8000/api/ipam/prefixes/1/available-prefixes/',
            params={},
            json=None,
            headers=HEADERS,
            verify=True
        )
        self.assertTrue(ret)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.post',
        return_value=Response(fixture='ipam/available-prefixes-post.json')
    )
    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='ipam/prefix.json'),
    )
    def test_create_available_prefixes(self, _, post):
        create_parms = dict(
            prefix_length=30,
        )
        pfx = nb.prefixes.get(1)
        ret = pfx.available_prefixes.create(create_parms)
        post.assert_called_with(
            'http://localhost:8000/api/ipam/prefixes/1/available-prefixes/',
            params={},
            headers=POST_HEADERS,
            json=create_parms,
            verify=True
        )
        self.assertTrue(ret)


class IPAddressTestCase(Generic.Tests):
    name = 'ip_addresses'
    name_singular = 'ip_address'

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='ipam/ip_address.json')
    )
    def test_modify(self, *_):
        ret = nb.prefixes.get(1)
        ret.description = 'testing'
        ret_serialized = ret.serialize()
        self.assertTrue(ret_serialized)
        self.assertFalse(ret._compare())
        self.assertEqual(ret_serialized['address'], '10.0.255.1/32')
        self.assertEqual(ret_serialized['description'], 'testing')


class RoleTestCase(Generic.Tests):
    name = 'roles'


class RIRTestCase(Generic.Tests):
    name = 'rirs'


class AggregatesTestCase(Generic.Tests):
    name = 'aggregates'


class VlanTestCase(Generic.Tests):
    name = 'vlans'


class VlanGroupsTestCase(Generic.Tests):
    name = 'vlan_groups'


class VRFTestCase(Generic.Tests):
    name = 'vrfs'


# class ServicesTestCase(Generic.Tests):
#     name = 'services'



import unittest
import six

import pynetbox
from .util import Response

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


api = pynetbox.api(
    "http://localhost:8000",
    token="abc123",
)

nb = api.dcim

HEADERS = {
    'accept': 'application/json;',
    'authorization': 'Token abc123',
}


class Generic(object):
    class Tests(unittest.TestCase):
        name = ''
        ret = pynetbox.core.response.Record
        app = 'dcim'

        def test_get_all(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name
                ))
            ) as mock:
                ret = getattr(nb, self.name).all()
                self.assertTrue(ret)
                self.assertTrue(isinstance(ret, list))
                self.assertTrue(isinstance(ret[0], self.ret))
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )

        def test_filter(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name
                ))
            ) as mock:
                ret = getattr(nb, self.name).filter(name='test')
                self.assertTrue(ret)
                self.assertTrue(isinstance(ret, list))
                self.assertTrue(isinstance(ret[0], self.ret))
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={"name": "test"},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )

        def test_get(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name[:-1]
                ))
            ) as mock:
                ret = getattr(nb, self.name).get(1)
                self.assertTrue(ret)
                self.assertTrue(isinstance(ret, self.ret))
                self.assertTrue(isinstance(str(ret), str))
                self.assertTrue(isinstance(dict(ret), dict))
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/1/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )

        def test_delete(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name[:-1]
                ))
            ) as mock, patch('pynetbox.core.query.requests.sessions.Session.delete') as delete:
                ret = getattr(nb, self.name).get(1)
                self.assertTrue(ret.delete())
                mock.assert_called_with(
                    'http://localhost:8000/api/{}/{}/1/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )
                delete.assert_called_with(
                    'http://localhost:8000/api/{}/{}/1/'.format(
                        self.app,
                        self.name.replace('_', '-')
                    ),
                    params={},
                    json=None,
                    headers=HEADERS,
                    verify=True
                )

        def test_compare(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name[:-1]
                ))
            ):
                ret = getattr(nb, self.name).get(1)
                self.assertTrue(ret)
                self.assertTrue(ret._compare())

        def test_serialize(self):
            with patch(
                'pynetbox.core.query.requests.sessions.Session.get',
                return_value=Response(fixture='{}/{}.json'.format(
                    self.app,
                    self.name[:-1]
                ))
            ):
                ret = getattr(nb, self.name).get(1)
                self.assertTrue(ret)
                self.assertTrue(ret.serialize())


class DeviceTestCase(Generic.Tests):
    name = 'devices'

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='dcim/device.json')
    )
    def test_get(self, mock):
        ret = getattr(nb, self.name).get(1)
        self.assertTrue(ret)
        self.assertTrue(isinstance(ret, self.ret))
        self.assertTrue(isinstance(ret.primary_ip, pynetbox.models.ipam.IpAddresses))
        self.assertTrue(isinstance(ret.primary_ip4, pynetbox.models.ipam.IpAddresses))
        self.assertTrue(isinstance(ret.config_context, dict))
        self.assertTrue(isinstance(ret.custom_fields, dict))
        self.assertTrue(isinstance(ret.local_context_data, dict))
        mock.assert_called_with(
            'http://localhost:8000/api/{}/{}/1/'.format(
                self.app,
                self.name.replace('_', '-')
            ),
            params={},
            json=None,
            headers=HEADERS,
            verify=True
        )

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='dcim/devices.json')
    )
    def test_multi_filter(self, mock):
        ret = getattr(nb, self.name).filter(role=['test', 'test1'], site='TEST#1')
        self.assertTrue(ret)
        self.assertTrue(isinstance(ret, list))
        self.assertTrue(isinstance(ret[0], self.ret))
        mock.assert_called_with(
            'http://localhost:8000/api/{}/{}/'.format(
                self.app,
                self.name.replace('_', '-')
            ),
            params={'role': ['test', 'test1'], 'site': 'TEST#1'},
            json=None,
            headers=HEADERS,
            verify=True
        )

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='dcim/device.json')
    )
    def test_modify(self, *_):
        ret = nb.devices.get(1)
        ret.serial = '123123123123'
        ret_serialized = ret.serialize()
        self.assertTrue(ret_serialized)
        self.assertFalse(ret._compare())
        self.assertEqual(ret_serialized['serial'], '123123123123')

    @patch(
        'pynetbox.core.query.requests.sessions.Session.post',
        return_value=Response(fixture='dcim/device.json')
    )
    def test_create(self, *_):
        data = {
            'name': 'test-device',
            'site': 1,
            'device_type': 1,
            'device_role': 1,
        }
        ret = nb.devices.create(**data)
        self.assertTrue(ret)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.post',
        return_value=Response(fixture='dcim/device_bulk_create.json')
    )
    def test_create_device_bulk(self, *_):
        data = [
            {
                'name': 'test-device',
                'site': 1,
                'device_type': 1,
                'device_role': 1,
            },
            {
                'name': 'test-device1',
                'site': 1,
                'device_type': 1,
                'device_role': 1,
            },
        ]
        ret = nb.devices.create(data)
        self.assertTrue(ret)
        self.assertTrue(len(ret), 2)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        side_effect=[
            Response(fixture='dcim/device.json'),
            Response(fixture='dcim/rack.json'),
        ]
    )
    def test_get_recurse(self, *_):
        '''Test that automatic recursion works, and that nested items
        are converted to Response() objects.
        '''
        ret = getattr(nb, self.name).get(1)
        self.assertTrue(ret)
        self.assertTrue(isinstance(ret, self.ret))
        self.assertTrue(isinstance(
            ret.rack.role,
            self.ret
        ))

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        side_effect=[
            Response(fixture='dcim/device.json'),
            Response(fixture='dcim/napalm.json'),
        ]
    )
    def test_get_napalm(self, mock):
        test = nb.devices.get(1)
        ret = test.napalm.list(method='get_facts')
        mock.assert_called_with(
            'http://localhost:8000/api/dcim/devices/1/napalm/',
            params={"method": "get_facts"},
            json=None,
            headers=HEADERS,
            verify=True,
        )
        self.assertTrue(ret)
        self.assertTrue(ret['get_facts'])


class SiteTestCase(Generic.Tests):
    name = 'sites'

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='dcim/site.json')
    )
    def test_modify_custom(self, *_):
        '''Test modifying a custom field.
        '''
        ret = getattr(nb, self.name).get(1)
        ret.custom_fields['test_custom'] = "Testing"
        self.assertFalse(ret._compare())
        self.assertTrue(ret.serialize())
        self.assertEqual(
            ret.custom_fields['test_custom'], 'Testing'
        )

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='dcim/site.json')
    )
    def test_custom_selection_serializer(self, _):
        '''Tests serializer with custom selection fields.
        '''
        ret = getattr(nb, self.name).get(1)
        ret.custom_fields['test_custom'] = "Testing"
        test = ret.serialize()
        self.assertEqual(
            test['custom_fields']['test_selection'],
            2
        )

    @patch(
        'pynetbox.core.query.requests.sessions.Session.post',
        return_value=Response(fixture='dcim/site.json')
    )
    def test_create(self, *_):
        data = {
            'name': 'TEST1',
            'custom_fields': {
                'test_custom': 'Testing'
            }
        }
        ret = nb.sites.create(**data)
        self.assertTrue(ret)


class InterfaceTestCase(Generic.Tests):
    name = 'interfaces'

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        return_value=Response(fixture='dcim/interface.json')
    )
    def test_modify(self, *_):
        ret = nb.interfaces.get(1)
        ret.description = 'Testing'
        ret_serialized = ret.serialize()
        self.assertTrue(ret)
        self.assertEqual(ret_serialized['description'], 'Testing')
        self.assertEqual(ret_serialized['form_factor'], 1400)

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        side_effect=[
            Response(fixture='dcim/{}.json'.format(name + '_1')),
            Response(fixture='dcim/{}.json'.format(name + '_2')),
        ]
    )
    def test_get_all(self, mock):
        ret = getattr(nb, self.name).all()
        self.assertTrue(ret)
        self.assertTrue(isinstance(ret, list))
        self.assertTrue(isinstance(ret[0], self.ret))
        self.assertEqual(len(ret), 71)
        mock.assert_called_with(
            'http://localhost:8000/api/dcim/interfaces/',
            params={"limit": 221, "offset": 50},
            json=None,
            headers=HEADERS,
            verify=True
        )


class RackTestCase(Generic.Tests):
    name = 'racks'

    @patch(
        'pynetbox.core.query.requests.sessions.Session.get',
        side_effect=[
            Response(fixture='dcim/rack.json'),
            Response(fixture='dcim/rack_u.json'),
        ]
    )
    def test_get_units(self, mock):
        test = nb.racks.get(1)
        ret = test.units.list()
        mock.assert_called_with(
            'http://localhost:8000/api/dcim/racks/1/units/',
            params={},
            json=None,
            headers=HEADERS,
            verify=True,
        )
        self.assertTrue(ret)
        self.assertTrue(
            isinstance(ret[0].device, pynetbox.models.dcim.Devices)
        )


class RackRoleTestCase(Generic.Tests):
    name = 'rack_roles'


class RegionTestCase(Generic.Tests):
    name = 'regions'


class RackGroupsTestCase(Generic.Tests):
    name = 'rack_groups'


class RackReservationsTestCase(Generic.Tests):
    name = 'rack_reservations'


class ManufacturersTestCase(Generic.Tests):
    name = 'manufacturers'


class DeviceTypeTestCase(Generic.Tests):
    name = 'device_types'


class ConsolePortTemplateTestCase(Generic.Tests):
    name = 'console_port_templates'


class ConsoleServerPortTemplateTestCase(Generic.Tests):
    name = 'console_server_port_templates'


class PowerPortTemplateTestCase(Generic.Tests):
    name = 'power_port_templates'


class PowerOutletTemplateTestCase(Generic.Tests):
    name = 'power_outlet_templates'


class InterfaceTemplateTestCase(Generic.Tests):
    name = 'interface_templates'


class DeviceBayTemplateTestCase(Generic.Tests):
    name = 'device_bay_templates'


class DeviceRolesTestCase(Generic.Tests):
    name = 'device_roles'


class PlatformsTestCase(Generic.Tests):
    name = 'platforms'


class ConsolePortsTestCase(Generic.Tests):
    name = 'console_ports'


class ConsoleServerPortsTestCase(Generic.Tests):
    name = 'console_server_ports'


class PowerPortsTestCase(Generic.Tests):
    name = 'power_ports'


class PowerOutletsTestCase(Generic.Tests):
    name = 'power_outlets'


class DeviceBaysTestCase(Generic.Tests):
    name = 'device_bays'


# class InventoryItemsTestCase(Generic.Tests):
#     name = 'inventory_items'


class InterfaceConnectionsTestCase(Generic.Tests):
    name = 'interface_connections'


# class ConnectedDevicesTestCase(Generic.Tests):
#     name = 'connected_device'


class VirtualChassisTestCase(Generic.Tests):
    name = 'virtual_chassis_devices'


class Choices(unittest.TestCase):

    def test_get(self):
        with patch(
            'pynetbox.core.query.requests.sessions.Session.get',
            return_value=Response(fixture='{}/{}.json'.format(
                'dcim',
                'choices'
            ))
        ) as mock:
            ret = nb.choices()
            self.assertTrue(ret)
            mock.assert_called_with(
                'http://localhost:8000/api/dcim/_choices/',
                params={},
                json=None,
                headers=HEADERS,
                verify=True
            )


class CablesTestCase(Generic.Tests):
    name = 'cables'

    def test_get_circuit(self):
        response_obj = Response(content={
            "id": 1,
            "termination_a_type": "circuits.circuittermination",
            "termination_a_id": 1,
            "termination_a": {
                "id": 1,
                "url": "http://localhost:8000/api/circuits/circuit-terminations/1/",
                "circuit": {
                    "id": 346,
                    "url": "http://localhost:8000/api/circuits/circuits/1/",
                    "cid": "TEST123321"
                },
                "term_side": "A",
            },
            "termination_b_type": "dcim.interface",
            "termination_b_id": 2,
            "termination_b": {
                "id": 2,
                "url": "http://localhost:8000/api/dcim/interfaces/2/",
                "device": {
                    "id": 2,
                    "url": "http://localhost:8000/api/dcim/devices/2/",
                    "name": "tst1-test2",
                    "display_name": "tst1-test2"
                },
                "name": "xe-0/0/0",
                "cable": 1
            },
            "type": None,
            "status": {
                "value": True,
                "label": "Connected"
            },
            "label": "",
            "color": "",
            "length": None,
            "length_unit": None
        })
        with patch(
            'pynetbox.core.query.requests.sessions.Session.get',
            return_value=response_obj
        ) as mock:
            ret = getattr(nb, self.name).get(1)
            self.assertTrue(ret)
            self.assertTrue(isinstance(ret, self.ret))
            self.assertTrue(isinstance(str(ret), str))
            self.assertTrue(isinstance(dict(ret), dict))
            mock.assert_called_with(
                'http://localhost:8000/api/{}/{}/1/'.format(
                    self.app,
                    self.name.replace('_', '-')
                ),
                headers=HEADERS,
                params={},
                json=None,
                verify=True
            )



import unittest

import six

from pynetbox.core.response import Record

if six.PY3:
    from unittest.mock import Mock
else:
    from mock import Mock


class RecordTestCase(unittest.TestCase):
    def test_serialize_list_of_records(self):
        test_values = {
            "id": 123,
            "tagged_vlans": [
                {
                    "id": 1,
                    "url": "http://localhost:8000/api/ipam/vlans/1/",
                    "vid": 1,
                    "name": "test1",
                    "display_name": "test1",
                },
                {
                    "id": 2,
                    "url": "http://localhost:8000/api/ipam/vlans/2/",
                    "vid": 2,
                    "name": "test 2",
                    "display_name": "test2",
                },
            ],
        }
        test_obj = Record(test_values, None, None)
        test = test_obj.serialize()
        self.assertEqual(test["tagged_vlans"], [1, 2])

    def test_serialize_list_of_ints(self):
        test_values = {"id": 123, "units": [12]}
        test_obj = Record(test_values, None, None)
        test = test_obj.serialize()
        self.assertEqual(test["units"], [12])

    def test_serialize_tag_set(self):
        test_values = {"id": 123, "tags": ["foo", "bar", "foo"]}
        test = Record(test_values, None, None).serialize()
        self.assertEqual(len(test["tags"]), 2)

    def test_diff(self):
        test_values = {
            "id": 123,
            "custom_fields": {"foo": "bar"},
            "string_field": "foobar",
            "int_field": 1,
            "nested_dict": {"id": 222, "name": "bar"},
            "tags": ["foo", "bar"],
            "int_list": [123, 321, 231],
        }
        test = Record(test_values, None, None)
        test.tags.append("baz")
        test.nested_dict = 1
        test.string_field = "foobaz"
        self.assertEqual(test._diff(), {"tags", "nested_dict", "string_field"})

    def test_diff_append_records_list(self):
        test_values = {
            "id": 123,
            "tagged_vlans": [
                {
                    "id": 1,
                    "url": "http://localhost:8000/api/ipam/vlans/1/",
                    "vid": 1,
                    "name": "test1",
                    "display_name": "test1",
                }
            ],
        }
        test_obj = Record(test_values, None, None)
        test_obj.tagged_vlans.append(1)
        test = test_obj._diff()
        self.assertFalse(test)

    def test_dict(self):
        test_values = {
            "id": 123,
            "custom_fields": {"foo": "bar"},
            "string_field": "foobar",
            "int_field": 1,
            "nested_dict": {"id": 222, "name": "bar"},
            "tags": ["foo", "bar"],
            "int_list": [123, 321, 231],
            "empty_list": [],
            "record_list": [
                {
                    "id": 123,
                    "name": "Test",
                    "str_attr": "foo",
                    "int_attr": 123,
                    "custom_fields": {"foo": "bar"},
                    "tags": ["foo", "bar"],
                },
                {
                    "id": 321,
                    "name": "Test 1",
                    "str_attr": "bar",
                    "int_attr": 321,
                    "custom_fields": {"foo": "bar"},
                    "tags": ["foo", "bar"],
                },
            ],
        }
        test = Record(test_values, None, None)
        self.assertEqual(dict(test), test_values)

    def test_choices_idempotence_prev27(self):
        test_values = {
            "id": 123,
            "choices_test": {
                "value": 1,
                "label": "test",
            },
        }
        test = Record(test_values, None, None)
        test.choices_test = 1
        self.assertFalse(test._diff())

    def test_choices_idempotence_v27(self):
        test_values = {
            "id": 123,
            "choices_test": {
                "value": "test",
                "label": "test",
                "id": 1,
            },
        }
        test = Record(test_values, None, None)
        test.choices_test = "test"
        self.assertFalse(test._diff())

    def test_choices_idempotence_v28(self):
        test_values = {
            "id": 123,
            "choices_test": {
                "value": "test",
                "label": "test",
            },
        }
        test = Record(test_values, None, None)
        test.choices_test = "test"
        self.assertFalse(test._diff())

    def test_hash(self):
        endpoint = Mock()
        endpoint.name = "test-endpoint"
        test = Record({}, None, endpoint)
        self.assertIsInstance(hash(test), int)

    def test_hash_diff(self):
        endpoint1 = Mock()
        endpoint1.name = "test-endpoint"
        endpoint2 = Mock()
        endpoint2.name = "test-endpoint"
        test1 = Record({}, None, endpoint1)
        test1.id = 1
        test2 = Record({}, None, endpoint2)
        test2.id = 2
        self.assertNotEqual(hash(test1), hash(test2))

    def test_compare(self):
        endpoint1 = Mock()
        endpoint1.name = "test-endpoint"
        endpoint2 = Mock()
        endpoint2.name = "test-endpoint"
        test1 = Record({}, None, endpoint1)
        test1.id = 1
        test2 = Record({}, None, endpoint2)
        test2.id = 1
        self.assertEqual(test1, test2)

    def test_nested_write(self):
        app = Mock()
        app.token = 'abc123'
        endpoint = Mock()
        endpoint.name = "test-endpoint"
        test = Record({
            'id': 123,
            'name': 'test',
            'child': {
                'id': 321,
                'name': 'test123',
                'url': 'http://localhost:8080/test',
            },
        }, app, endpoint)
        test.child.name = 'test321'
        test.child.save()
        self.assertEqual(app.http_session.patch.call_args[0][0], "http://localhost:8080/test/")



import unittest
from collections import OrderedDict

from pynetbox.core.query import url_param_builder


class TestQuery(unittest.TestCase):

    def test_url_param_builder(self):
        test_params = OrderedDict([
            ('abc', '123'),
            ('xyz', '321'),
            ('enc', '#12'),
        ])
        self.assertEqual(url_param_builder(test_params), '?abc=123&xyz=321&enc=%2312')



import unittest

import six

from pynetbox.core.endpoint import Endpoint

if six.PY3:
    from unittest.mock import patch, Mock, call
else:
    from mock import patch, Mock, call


class EndPointTestCase(unittest.TestCase):

    def test_filter(self):
        with patch(
            "pynetbox.core.query.Request.get", return_value=Mock()
        ) as mock:
            api = Mock(base_url="http://localhost:8000/api")
            app = Mock(name="test")
            mock.return_value = [{'id': 123}, {'id': 321}]
            test_obj = Endpoint(api, app, "test")
            test = test_obj.filter(test="test")
            self.assertEqual(len(test), 2)

    def test_filter_empty_kwargs(self):

        api = Mock(base_url="http://localhost:8000/api")
        app = Mock(name="test")
        test_obj = Endpoint(api, app, "test")
        with self.assertRaises(ValueError) as _:
            test_obj.filter()

    def test_filter_reserved_kwargs(self):

        api = Mock(base_url="http://localhost:8000/api")
        app = Mock(name="test")
        test_obj = Endpoint(api, app, "test")
        with self.assertRaises(ValueError) as _:
            test_obj.filter(id=1)

    def test_choices(self):
        with patch(
            "pynetbox.core.query.Request.options", return_value=Mock()
        ) as mock:
            api = Mock(base_url="http://localhost:8000/api")
            app = Mock(name="test")
            mock.return_value = {
                "actions": {
                    "POST": {
                        "letter": {
                            "choices": [
                                {"display_name": "A", "value": 1},
                                {"display_name": "B", "value": 2},
                                {"display_name": "C", "value": 3},
                            ]
                        }
                    }
                }
            }
            test_obj = Endpoint(api, app, "test")
            choices = test_obj.choices()
            self.assertEqual(choices["letter"][1]["display_name"], "B")
            self.assertEqual(choices["letter"][1]["value"], 2)



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from pynetbox.core.query import Request
from pynetbox.core.util import Hashabledict


# List of fields that are lists but should be treated as sets.
LIST_AS_SET = ("tags", "tagged_vlans")


def get_return(lookup, return_fields=None):
    """Returns simple representations for items passed to lookup.

    Used to return a "simple" representation of objects and collections
    sent to it via lookup. Otherwise, we look to see if
    lookup is a "choices" field (dict with only 'id' and 'value')
    or a nested_return. Finally, we check if it's a Record, if
    so simply return a string. Order is important due to nested_return
    being self-referential.

    :arg list,optional return_fields: A list of fields to reference when
        calling values on lookup.
    """

    for i in return_fields or ["id", "value", "nested_return"]:
        if isinstance(lookup, dict) and lookup.get(i):
            return lookup[i]
        else:
            if hasattr(lookup, i):
                # check if this is a "choices" field record
                # from a NetBox 2.7 server.
                if sorted(dict(lookup)) == sorted(["id", "value", "label"]):
                    return getattr(lookup, "value")
                return getattr(lookup, i)

    if isinstance(lookup, Record):
        return str(lookup)
    else:
        return lookup


def flatten_custom(custom_dict):
    return {
        k: v if not isinstance(v, dict) else v["value"]
        for k, v in custom_dict.items()
    }


class JsonField(object):
    """Explicit field type for values that are not to be converted
    to a Record object"""
    _json_field = True


class Record(object):
    """Create python objects from netbox API responses.

    Creates an object from a NetBox response passed as `values`.
    Nested dicts that represent other endpoints are also turned
    into Record objects. All fields are then assigned to the
    object's attributes. If a missing attr is requested
    (e.g. requesting a field that's only present on a full response on
    a Record made from a nested response) the pynetbox will make a
    request for the full object and return the requsted value.

    :examples:

    Default representation of the object is usually its name

    >>> x = nb.dcim.devices.get(1)
    >>> x
    test1-switch1
    >>>

    Querying a string field.

    >>> x = nb.dcim.devices.get(1)
    >>> x.serial
    'ABC123'
    >>>

    Querying a field on a nested object.

    >>> x = nb.dcim.devices.get(1)
    >>> x.device_type.model
    'QFX5100-24Q'
    >>>

    Casting the object as a dictionary.

    >>> from pprint import pprint
    >>> pprint(dict(x))
    {'asset_tag': None,
     'cluster': None,
     'comments': '',
     'config_context': {},
     'created': '2018-04-01',
     'custom_fields': {},
     'device_role': {'id': 1,
                     'name': 'Test Switch',
                     'slug': 'test-switch',
                     'url': 'http://localhost:8000/api/dcim/device-roles/1/'},
     'device_type': {...},
     'display_name': 'test1-switch1',
     'face': {'label': 'Rear', 'value': 1},
     'id': 1,
     'name': 'test1-switch1',
     'parent_device': None,
     'platform': {...},
     'position': 1,
     'primary_ip': {'address': '192.0.2.1/24',
                    'family': 4,
                    'id': 1,
                    'url': 'http://localhost:8000/api/ipam/ip-addresses/1/'},
     'primary_ip4': {...},
     'primary_ip6': None,
     'rack': {'display_name': 'Test Rack',
              'id': 1,
              'name': 'Test Rack',
              'url': 'http://localhost:8000/api/dcim/racks/1/'},
     'serial': 'ABC123',
     'site': {'id': 1,
              'name': 'TEST',
              'slug': 'TEST',
              'url': 'http://localhost:8000/api/dcim/sites/1/'},
     'status': {'label': 'Active', 'value': 1},
     'tags': [],
     'tenant': None,
     'vc_position': None,
     'vc_priority': None,
     'virtual_chassis': None}
     >>>

     Iterating over a Record object.

    >>> for i in x:
    ...  print(i)
    ...
    ('id', 1)
    ('name', 'test1-switch1')
    ('display_name', 'test1-switch1')
    >>>

    """

    url = None

    def __init__(self, values, api, endpoint):
        self.has_details = False
        self._full_cache = []
        self._init_cache = []
        self.api = api
        self.endpoint = endpoint
        self.default_ret = Record

        if values:
            self._parse_values(values)

    def __getattr__(self, k):
        """Default behavior for missing attrs.

        We'll call `full_details()` if we're asked for an attribute
        we don't have.

        In order to prevent non-explicit behavior,`k='keys'` is
        excluded because casting to dict() calls this attr.
        """
        if self.url:
            if self.has_details is False and k != "keys":
                if self.full_details():
                    ret = getattr(self, k, None)
                    if ret or hasattr(self, k):
                        return ret

        raise AttributeError('object has no attribute "{}"'.format(k))

    def __iter__(self):
        for i in dict(self._init_cache):
            cur_attr = getattr(self, i)
            if isinstance(cur_attr, Record):
                yield i, dict(cur_attr)
            elif isinstance(cur_attr, list) and all(
                isinstance(i, Record) for i in cur_attr
            ):
                yield i, [dict(x) for x in cur_attr]
            else:
                yield i, cur_attr

    def __getitem__(self, item):
        return item

    def __str__(self):
        return (
            getattr(self, "name", None) or getattr(self, "label", None) or ""
        )

    def __repr__(self):
        return str(self)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __key__(self):
        if hasattr(self, "id"):
            return (self.endpoint.name, self.id)
        else:
            return (self.endpoint.name)

    def __hash__(self):
        return hash(self.__key__())

    def __eq__(self, other):
        if isinstance(other, Record):
            return self.__key__() == other.__key__()
        return NotImplemented

    def _add_cache(self, item):
        key, value = item
        self._init_cache.append((key, get_return(value)))

    def _parse_values(self, values):
        """ Parses values init arg.

        Parses values dict at init and sets object attributes with the
        values within.
        """

        def list_parser(list_item):
            if isinstance(list_item, dict):
                return self.default_ret(list_item, self.api, self.endpoint)
            return list_item

        for k, v in values.items():
            if isinstance(v, dict):
                lookup = getattr(self.__class__, k, None)
                if k == "custom_fields" or hasattr(lookup, "_json_field"):
                    self._add_cache((k, v.copy()))
                    setattr(self, k, v)
                    continue
                if lookup:
                    v = lookup(v, self.api, self.endpoint)
                else:
                    v = self.default_ret(v, self.api, self.endpoint)
                self._add_cache((k, v))

            elif isinstance(v, list):
                v = [list_parser(i) for i in v]
                to_cache = list(v)
                self._add_cache((k, to_cache))

            else:
                self._add_cache((k, v))
            setattr(self, k, v)

    def _compare(self):
        """Compares current attributes to values at instantiation.

        In order to be idempotent we run this method in `save()`.

        Returns:
            Boolean value, True indicates current instance has the same
            attributes as the ones passed to `values`.
        """

        if self.serialize(init=True) == self.serialize():
            return True
        return False

    def full_details(self):
        """Queries the hyperlinked endpoint if 'url' is defined.

        This method will populate the attributes from the detail
        endpoint when it's called. Sets the class-level `has_details`
        attribute when it's called to prevent being called more
        than once.

        :returns: True
        """
        if self.url:
            req = Request(
                base=self.url,
                token=self.api.token,
                session_key=self.api.session_key,
                ssl_verify=self.api.ssl_verify,
                http_session=self.api.http_session,
            )
            self._parse_values(req.get())
            self.has_details = True
            return True
        return False

    def serialize(self, nested=False, init=False):
        """Serializes an object

        Pulls all the attributes in an object and creates a dict that
        can be turned into the json that netbox is expecting.

        If an attribute's value is a ``Record`` type it's replaced with
        the ``id`` field of that object.

        .. note::

            Using this to get a dictionary representation of the record
            is discouraged. It's probably better to cast to dict()
            instead. See Record docstring for example.

        :returns: dict.
        """
        if nested:
            return get_return(self)

        if init:
            init_vals = dict(self._init_cache)

        ret = {}
        for i in dict(self):
            current_val = getattr(self, i) if not init else init_vals.get(i)
            if i == "custom_fields":
                ret[i] = flatten_custom(current_val)
            else:
                if isinstance(current_val, Record):
                    current_val = getattr(current_val, "serialize")(
                        nested=True
                    )

                if isinstance(current_val, list):
                    current_val = [
                        v.id if isinstance(v, Record) else v
                        for v in current_val
                    ]
                    if i in LIST_AS_SET:
                        current_val = list(set(current_val))
                ret[i] = current_val
        return ret

    def _diff(self):
        def fmt_dict(k, v):
            if isinstance(v, dict):
                return k, Hashabledict(v)
            if isinstance(v, list):
                return k, ",".join(map(str, v))
            return k, v

        current = Hashabledict(
            {fmt_dict(k, v) for k, v in self.serialize().items()}
        )
        init = Hashabledict(
            {fmt_dict(k, v) for k, v in self.serialize(init=True).items()}
        )
        return set([i[0] for i in set(current.items()) ^ set(init.items())])

    def save(self):
        """Saves changes to an existing object.

        Takes a diff between the objects current state and its state at init
        and sends them as a dictionary to Request.patch().

        :returns: True if PATCH request was successful.
        :example:

        >>> x = nb.dcim.devices.get(name='test1-a3-tor1b')
        >>> x.serial
        u''
        >>> x.serial = '1234'
        >>> x.save()
        True
        >>>
        """
        if self.id:
            diff = self._diff()
            if diff:
                serialized = self.serialize()
                req = Request(
                    key=self.id if not self.url else None,
                    base=self.url or self.endpoint.url,
                    token=self.api.token,
                    session_key=self.api.session_key,
                    ssl_verify=self.api.ssl_verify,
                    http_session=self.api.http_session,
                )
                if req.patch({i: serialized[i] for i in diff}):
                    return True

        return False

    def update(self, data):
        """Update an object with a dictionary.

        Accepts a dict and uses it to update the record and call save().
        For nested and choice fields you'd pass an int the same as
        if you were modifying the attribute and calling save().

        :arg dict data: Dictionary containing the k/v to update the
            record object with.
        :returns: True if PATCH request was successful.
        :example:

        >>> x = nb.dcim.devices.get(1)
        >>> x.update({
        ...   "name": "test-switch2",
        ...   "serial": "ABC321",
        ... })
        True

        """

        for k, v in data.items():
            setattr(self, k, v)
        return self.save()

    def delete(self):
        """Deletes an existing object.

        :returns: True if DELETE operation was successful.
        :example:

        >>> x = nb.dcim.devices.get(name='test1-a3-tor1b')
        >>> x.delete()
        True
        >>>
        """
        req = Request(
            key=self.id if not self.url else None,
            base=self.url or self.endpoint.url,
            token=self.api.token,
            session_key=self.api.session_key,
            ssl_verify=self.api.ssl_verify,
            http_session=self.api.http_session,
        )
        return True if req.delete() else False



import json


class Response(object):
    def __init__(self, fixture=None, status_code=200, ok=True, content=None):
        self.status_code = status_code
        self.content = json.dumps(content) if content else self.load_fixture(fixture)
        self.ok = ok

    def load_fixture(self, path):
        with open("tests/fixtures/{}".format(path), 'r') as f:
            return f.read()

    def json(self):
        return json.loads(self.content)



import unittest

import six

import pynetbox

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch

host = "http://localhost:8000"

def_kwargs = {
    'token': 'abc123',
}


class AppCustomChoicesTestCase(unittest.TestCase):

    @patch(
        'pynetbox.core.query.Request.get',
        return_value={
            "Testfield1": {"TF1_1": 1, "TF1_2": 2},
            "Testfield2": {"TF2_1": 3, "TF2_2": 4},
        }
    )
    def test_custom_choices(self, *_):
        api = pynetbox.api(
            host,
            **def_kwargs
        )
        choices = api.extras.custom_choices()
        self.assertEqual(len(choices), 2)
        self.assertEqual(sorted(choices.keys()), ["Testfield1", "Testfield2"])



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from pynetbox.core.response import Record
from pynetbox.core.endpoint import DetailEndpoint


class IpAddresses(Record):
    def __str__(self):
        return str(self.address)


class Prefixes(Record):
    def __str__(self):
        return str(self.prefix)

    @property
    def available_ips(self):
        """ Represents the ``available-ips`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing and creating IP addresses inside a prefix.

        :returns: :py:class:`.DetailEndpoint`

        :Examples:

        >>> prefix = nb.ipam.prefixes.get(24)
        >>> prefix.available_ips.list()
        [{u'vrf': None, u'family': 4, u'address': u'10.1.1.49/30'}...]

        To create a single IP:

        >>> prefix = nb.ipam.prefixes.get(24)
        >>> prefix.available_ips.create()
        {u'status': 1, u'description': u'', u'nat_inside': None...}


        To create multiple IPs:

        >>> prefix = nb.ipam.prefixes.get(24)
        >>> create = prefix.available_ips.create([{} for i in range(2)])
        >>> len(create)
        2
        """
        return DetailEndpoint(self, "available-ips")

    @property
    def available_prefixes(self):
        """ Represents the ``available-prefixes`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing and creating prefixes inside a parent prefix.

        Very similar to :py:meth:`~pynetbox.ipam.Prefixes.available_ips`
        , except that dict (or list of dicts) passed to ``.create()``
        needs to have a ``prefix_length`` key/value specifed.

        :returns: :py:class:`.DetailEndpoint`

        :Examples:

        >>> prefix.available_prefixes.list()
        [{u'prefix': u'10.1.1.44/30', u'vrf': None, u'family': 4}]


        Creating a single child prefix:

        >>> prefix = nb.ipam.prefixes.get(1)
        >>> new_prefix = prefix.available_prefixes.create(
        ...    {'prefix_length': 29}
        ...)
        >>> new_prefix['prefix']
        u'10.1.1.56/29'

        """
        return DetailEndpoint(self, "available-prefixes")


class Aggregates(Record):
    def __str__(self):
        return str(self.prefix)
